@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2017 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -46,7 +46,7 @@ static int cached_len=0;
 
 #ifndef OPENSSL_NO_DH
 DH *dh_params=NULL;
-int dh_needed=0;
+int dh_temp_params=0;
 #endif /* OPENSSL_NO_DH */
 
 /**************************************** prototypes */
@@ -54,7 +54,7 @@ int dh_needed=0;
 /* SNI */
 #ifndef OPENSSL_NO_TLSEXT
 NOEXPORT int servername_cb(SSL *, int *, void *);
-NOEXPORT int matches_wildcard(char *, char *);
+NOEXPORT int matches_wildcard(const char *, const char *);
 #endif
 
 /* DH/ECDH */
@@ -92,6 +92,8 @@ NOEXPORT int ui_retry();
 
 /* session callbacks */
 NOEXPORT int sess_new_cb(SSL *, SSL_SESSION *);
+NOEXPORT void new_chain(CLI *);
+NOEXPORT void session_cache_save(CLI *, SSL_SESSION *);
 NOEXPORT SSL_SESSION *sess_get_cb(SSL *,
 #if OPENSSL_VERSION_NUMBER>=0x10100000L
     const
@@ -124,10 +126,29 @@ typedef long SSL_OPTIONS_TYPE;
 
 int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
     /* create TLS context */
+#if OPENSSL_VERSION_NUMBER>=0x10100000L
+    if(section->option.client)
+        section->ctx=SSL_CTX_new(TLS_client_method());
+    else /* server mode */
+        section->ctx=SSL_CTX_new(TLS_server_method());
+    if(!SSL_CTX_set_min_proto_version(section->ctx,
+            section->min_proto_version)) {
+        s_log(LOG_ERR, "Failed to set the minimum protocol version 0x%X",
+            section->min_proto_version);
+        return 1; /* FAILED */
+    }
+    if(!SSL_CTX_set_max_proto_version(section->ctx,
+            section->max_proto_version)) {
+        s_log(LOG_ERR, "Failed to set the maximum protocol version 0x%X",
+            section->max_proto_version);
+        return 1; /* FAILED */
+    }
+#else /* OPENSSL_VERSION_NUMBER<0x10100000L */
     if(section->option.client)
         section->ctx=SSL_CTX_new(section->client_method);
     else /* server mode */
         section->ctx=SSL_CTX_new(section->server_method);
+#endif /* OPENSSL_VERSION_NUMBER<0x10100000L */
     if(!section->ctx) {
         sslerror("SSL_CTX_new");
         return 1; /* FAILED */
@@ -148,19 +169,29 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
         }
     }
 
-    /* options */
+    /* TLS options: configure the stunnel defaults first */
+    SSL_CTX_set_options(section->ctx, SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3);
+#ifdef SSL_OP_NO_COMPRESSION
+    /* we implemented a better way to disable compression if needed */
+    SSL_CTX_clear_options(section->ctx, SSL_OP_NO_COMPRESSION);
+#endif /* SSL_OP_NO_COMPRESSION */
+
+    /* TLS options: configure the user-specified values */
     SSL_CTX_set_options(section->ctx,
         (SSL_OPTIONS_TYPE)(section->ssl_options_set));
 #if OPENSSL_VERSION_NUMBER>=0x009080dfL
     SSL_CTX_clear_options(section->ctx,
         (SSL_OPTIONS_TYPE)(section->ssl_options_clear));
+#endif /* OpenSSL 0.9.8m or later */
+
+    /* TLS options: log the configured values */
+#if OPENSSL_VERSION_NUMBER>=0x009080dfL
     s_log(LOG_DEBUG, "TLS options: 0x%08lX (+0x%08lX, -0x%08lX)",
         SSL_CTX_get_options(section->ctx),
         section->ssl_options_set, section->ssl_options_clear);
 #else /* OpenSSL older than 0.9.8m */
     s_log(LOG_DEBUG, "TLS options: 0x%08lX (+0x%08lX)",
-        SSL_CTX_get_options(section->ctx),
-        section->ssl_options_set);
+        SSL_CTX_get_options(section->ctx), section->ssl_options_set);
 #endif /* OpenSSL 0.9.8m or later */
 
     /* initialize OpenSSL CONF options */
@@ -190,6 +221,8 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
             return 1; /* FAILED */
         }
     }
+    SSL_CTX_set_session_cache_mode(section->ctx,
+        SSL_SESS_CACHE_BOTH | SSL_SESS_CACHE_NO_INTERNAL_STORE);
     SSL_CTX_sess_set_cache_size(section->ctx, section->session_size);
     SSL_CTX_set_timeout(section->ctx, section->session_timeout);
     SSL_CTX_sess_set_new_cb(section->ctx, sess_new_cb);
@@ -210,7 +243,6 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
     /* initialize the DH/ECDH key agreement in the server mode */
     if(!section->option.client) {
 #ifndef OPENSSL_NO_TLSEXT
-        SSL_CTX_set_tlsext_servername_arg(section->ctx, section);
         SSL_CTX_set_tlsext_servername_callback(section->ctx, servername_cb);
 #endif /* OPENSSL_NO_TLSEXT */
 #ifndef OPENSSL_NO_DH
@@ -229,17 +261,16 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
 #ifndef OPENSSL_NO_TLSEXT
 
 NOEXPORT int servername_cb(SSL *ssl, int *ad, void *arg) {
-    SERVICE_OPTIONS *section=(SERVICE_OPTIONS *)arg;
     const char *servername=SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name);
+    CLI *c=SSL_get_ex_data(ssl, index_ssl_cli);
     SERVERNAME_LIST *list;
-    CLI *c;
-#ifdef USE_LIBWRAP
-    char *accepted_address;
-#endif /* USE_LIBWRAP */
 
     /* leave the alert type at SSL_AD_UNRECOGNIZED_NAME */
     (void)ad; /* squash the unused parameter warning */
-    if(!section->servername_list_head) {
+    (void)arg; /* squash the unused parameter warning */
+
+    /* handle trivial cases first */
+    if(!c->opt->servername_list_head) {
         s_log(LOG_DEBUG, "SNI: no virtual services defined");
         return SSL_TLSEXT_ERR_OK;
     }
@@ -247,26 +278,31 @@ NOEXPORT int servername_cb(SSL *ssl, int *ad, void *arg) {
         s_log(LOG_NOTICE, "SNI: no servername received");
         return SSL_TLSEXT_ERR_NOACK;
     }
-    s_log(LOG_INFO, "SNI: requested servername: %s", servername);
 
-    for(list=section->servername_list_head; list; list=list->next)
-        if(matches_wildcard((char *)servername, list->servername)) {
-            s_log(LOG_DEBUG, "SNI: matched pattern: %s", list->servername);
-            c=SSL_get_ex_data(ssl, index_ssl_cli);
-            c->opt=list->opt;
-            SSL_set_SSL_CTX(ssl, c->opt->ctx);
-            SSL_set_verify(ssl, SSL_CTX_get_verify_mode(c->opt->ctx),
-                SSL_CTX_get_verify_callback(c->opt->ctx));
-            s_log(LOG_NOTICE, "SNI: switched to service [%s]",
-                c->opt->servname);
+    /* find a matching section */
+    s_log(LOG_INFO, "SNI: requested servername: %s", servername);
+    for(list=c->opt->servername_list_head; list; list=list->next)
+        if(matches_wildcard(servername, list->servername))
+            break;
+    if(!list) {
+        s_log(LOG_ERR, "SNI: no pattern matched servername: %s", servername);
+        return SSL_TLSEXT_ERR_OK;
+    }
+    s_log(LOG_DEBUG, "SNI: matched pattern: %s", list->servername);
+
+    /* switch to the new section */
+#ifndef USE_FORK
+    service_up_ref(list->opt);
+    service_free(c->opt);
+#endif
+    c->opt=list->opt;
+    SSL_set_SSL_CTX(ssl, c->opt->ctx);
+    SSL_set_verify(ssl, SSL_CTX_get_verify_mode(c->opt->ctx),
+        SSL_CTX_get_verify_callback(c->opt->ctx));
+    s_log(LOG_NOTICE, "SNI: switched to service [%s]", c->opt->servname);
 #ifdef USE_LIBWRAP
-            accepted_address=s_ntop(&c->peer_addr, c->peer_addr_len);
-            libwrap_auth(c, accepted_address); /* retry on a service switch */
-            str_free(accepted_address);
+    libwrap_auth(c); /* retry on a service switch */
 #endif /* USE_LIBWRAP */
-            return SSL_TLSEXT_ERR_OK;
-        }
-    s_log(LOG_ERR, "SNI: no pattern matched servername: %s", servername);
     return SSL_TLSEXT_ERR_OK;
 }
 /* TLSEXT callback return codes:
@@ -275,18 +311,17 @@ NOEXPORT int servername_cb(SSL *ssl, int *ad, void *arg) {
  *  - SSL_TLSEXT_ERR_ALERT_FATAL
  *  - SSL_TLSEXT_ERR_NOACK */
 
-NOEXPORT int matches_wildcard(char *servername, char *pattern) {
-    ssize_t diff;
-
+NOEXPORT int matches_wildcard(const char *servername, const char *pattern) {
     if(!servername || !pattern)
         return 0;
     if(*pattern=='*') { /* wildcard comparison */
-        diff=(ssize_t)strlen(servername)-(ssize_t)strlen(++pattern);
+        ssize_t diff=(ssize_t)strlen(servername)-((ssize_t)strlen(pattern)-1);
         if(diff<0) /* pattern longer than servername */
             return 0;
-        servername+=diff;
+        return !strcasecmp(servername+diff, pattern+1);
+    } else { /* string comparison */
+        return !strcasecmp(servername, pattern);
     }
-    return !strcasecmp(servername, pattern);
 }
 
 #endif /* OPENSSL_NO_TLSEXT */
@@ -307,8 +342,9 @@ NOEXPORT int dh_init(SERVICE_OPTIONS *section) {
     char description[128];
     STACK_OF(SSL_CIPHER) *ciphers;
 
+    section->option.dh_temp_params=0; /* disable by default */
+
     /* check if DH is actually enabled for this section */
-    section->option.dh_needed=0;
     ciphers=SSL_CTX_get_ciphers(section->ctx);
     if(!ciphers)
         return 1; /* ERROR (unlikely) */
@@ -319,12 +355,15 @@ NOEXPORT int dh_init(SERVICE_OPTIONS *section) {
             description, sizeof description);
         /* s_log(LOG_INFO, "Ciphersuite: %s", description); */
         if(strstr(description, " Kx=DH")) {
-            section->option.dh_needed=1; /* update this context */
+            s_log(LOG_INFO, "DH initialization needed for %s",
+                SSL_CIPHER_get_name(sk_SSL_CIPHER_value(ciphers, i)));
             break;
         }
     }
-    if(!section->option.dh_needed) /* no DH ciphers found */
+    if(i==n) { /* no DH ciphers found */
+        s_log(LOG_INFO, "DH initialization not needed");
         return 0; /* OK */
+    }
 
     s_log(LOG_DEBUG, "DH initialization");
 #ifndef OPENSSL_NO_ENGINE
@@ -337,11 +376,11 @@ NOEXPORT int dh_init(SERVICE_OPTIONS *section) {
         DH_free(dh);
         return 0; /* OK */
     }
-    stunnel_read_lock(&stunnel_locks[LOCK_DH]);
+    CRYPTO_THREAD_read_lock(stunnel_locks[LOCK_DH]);
     SSL_CTX_set_tmp_dh(section->ctx, dh_params);
-    stunnel_read_unlock(&stunnel_locks[LOCK_DH]);
-    dh_needed=1; /* generate temporary DH parameters in cron */
-    section->option.dh_needed=1; /* update this context */
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_DH]);
+    dh_temp_params=1; /* generate temporary DH parameters in cron */
+    section->option.dh_temp_params=1; /* update this section in cron */
     s_log(LOG_INFO, "Using dynamic DH parameters");
     return 0; /* OK */
 }
@@ -561,33 +600,21 @@ NOEXPORT unsigned psk_server_callback(SSL *ssl, const char *identity,
     unsigned char *psk, unsigned max_psk_len) {
     CLI *c;
     PSK_KEYS *found;
-    size_t len;
 
     c=SSL_get_ex_data(ssl, index_ssl_cli);
     found=psk_find(&c->opt->psk_sorted, identity);
-    if(found) {
-        len=found->key_len;
-    } else {
-        s_log(LOG_ERR, "No key found for PSK identity \"%s\"", identity);
-        len=0;
+    if(!found) {
+        s_log(LOG_INFO, "PSK identity not found (session resumption?)");
+        return 0;
     }
-    if(len>max_psk_len) {
-        s_log(LOG_ERR, "PSK too long (%lu>%d bytes)",
-            (long unsigned)len, max_psk_len);
-        len=0;
+    if(found->key_len>max_psk_len) {
+        s_log(LOG_ERR, "PSK too long (%u>%u)", found->key_len, max_psk_len);
+        return 0;
     }
-    if(len) {
-        memcpy(psk, found->key_val, len);
-        s_log(LOG_NOTICE, "Key configured for PSK identity \"%s\"", identity);
-    } else { /* block identity probes if possible */
-        if(max_psk_len>=32 && RAND_bytes(psk, 32)>0) {
-            len=32; /* 256 random bits */
-            s_log(LOG_ERR, "Configured random PSK");
-        } else {
-            s_log(LOG_ERR, "Rejecting with unknown_psk_identity alert");
-        }
-    }
-    return (unsigned)len;
+    memcpy(psk, found->key_val, found->key_len);
+    s_log(LOG_NOTICE, "Key configured for PSK identity \"%s\"", identity);
+    c->flag.psk=1;
+    return found->key_len;
 }
 
 NOEXPORT int psk_compar(const void *a, const void *b) {
@@ -608,7 +635,7 @@ void psk_sort(PSK_TABLE *table, PSK_KEYS *head) {
         ++table->num;
     s_log(LOG_INFO, "PSK identities: %lu retrieved",
         (long unsigned)table->num);
-    table->val=str_alloc(table->num*sizeof(PSK_KEYS *));
+    table->val=str_alloc_detached(table->num*sizeof(PSK_KEYS *));
     for(curr=head, i=0; i<table->num; ++i) {
         table->val[i]=curr;
         curr=curr->next;
@@ -901,11 +928,116 @@ NOEXPORT int sess_new_cb(SSL *ssl, SSL_SESSION *sess) {
 
     s_log(LOG_DEBUG, "New session callback");
     c=SSL_get_ex_data(ssl, index_ssl_cli);
+
+    new_chain(c); /* new session -> we may have a new peer certificate chain */
+
+    if(c->opt->option.client)
+        session_cache_save(c, sess);
+    else /* SSL_SESS_CACHE_NO_INTERNAL_STORE prevented automatic caching */
+        SSL_CTX_add_session(SSL_get_SSL_CTX(ssl), sess);
     if(c->opt->option.sessiond)
         cache_new(ssl, sess);
+
+    print_session_id(sess);
+
     return 0; /* the OpenSSL's manual is really bad -> use the source here */
 }
 
+#if OPENSSL_VERSION_NUMBER<0x0090800fL
+NOEXPORT const unsigned char *SSL_SESSION_get_id(const SSL_SESSION *s,
+        unsigned int *len) {
+    if(len)
+        *len=s->session_id_length;
+    return (const unsigned char *)s->session_id;
+}
+#endif
+
+void print_session_id(SSL_SESSION *sess) {
+    const unsigned char *session_id;
+    unsigned int session_id_length;
+    char session_id_txt[2*SSL_MAX_SSL_SESSION_ID_LENGTH+1];
+
+    session_id=SSL_SESSION_get_id(sess, &session_id_length);
+    bin2hexstring(session_id, session_id_length,
+        session_id_txt, sizeof session_id_txt);
+    s_log(LOG_INFO, "Session id: %s", session_id_txt);
+}
+
+NOEXPORT void new_chain(CLI *c) {
+    BIO *bio;
+    int i, len;
+    X509 *peer_cert;
+    STACK_OF(X509) *sk;
+    char *chain;
+
+    if(c->opt->chain) /* already cached */
+        return; /* this race condition is safe to ignore */
+    bio=BIO_new(BIO_s_mem());
+    if(!bio)
+        return;
+    sk=SSL_get_peer_cert_chain(c->ssl);
+    for(i=0; sk && i<sk_X509_num(sk); i++) {
+        peer_cert=sk_X509_value(sk, i);
+        PEM_write_bio_X509(bio, peer_cert);
+    }
+    if(!sk || !c->opt->option.client) {
+        peer_cert=SSL_get_peer_certificate(c->ssl);
+        if(peer_cert) {
+            PEM_write_bio_X509(bio, peer_cert);
+            X509_free(peer_cert);
+        }
+    }
+    len=BIO_pending(bio);
+    if(len<=0) {
+        s_log(LOG_INFO, "No peer certificate received");
+        BIO_free(bio);
+        return;
+    }
+    /* prevent automatic deallocation of the cached value */
+    chain=str_alloc_detached((size_t)len+1);
+    len=BIO_read(bio, chain, len);
+    if(len<0) {
+        s_log(LOG_ERR, "BIO_read failed");
+        BIO_free(bio);
+        str_free(chain);
+        return;
+    }
+    chain[len]='\0';
+    BIO_free(bio);
+    c->opt->chain=chain; /* this race condition is safe to ignore */
+    ui_new_chain(c->opt->section_number);
+    s_log(LOG_DEBUG, "Peer certificate was cached (%d bytes)", len);
+}
+
+/* cache client sessions */
+NOEXPORT void session_cache_save(CLI *c, SSL_SESSION *sess) {
+    SSL_SESSION *old;
+
+#if OPENSSL_VERSION_NUMBER>=0x10100000L
+    SSL_SESSION_up_ref(sess);
+#else
+    sess=SSL_get1_session(c->ssl);
+#endif
+
+    CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_SESSION]);
+    if(c->opt->option.delayed_lookup) {
+        old=c->opt->session;
+        c->opt->session=sess;
+    } else { /* per-destination client cache */
+        if(c->opt->connect_session) {
+            old=c->opt->connect_session[c->idx];
+            c->opt->connect_session[c->idx]=sess;
+        } else {
+            s_log(LOG_ERR, "INTERNAL ERROR: Uninitialized client session cache");
+            old=NULL;
+        }
+    }
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_SESSION]);
+
+    if(old)
+        SSL_SESSION_free(old);
+}
+
 NOEXPORT SSL_SESSION *sess_get_cb(SSL *ssl,
 #if OPENSSL_VERSION_NUMBER>=0x10100000L
         const
@@ -948,12 +1080,7 @@ NOEXPORT void cache_new(SSL *ssl, SSL_SESSION *sess) {
     val_tmp=val=str_alloc((size_t)val_len);
     i2d_SSL_SESSION(sess, &val_tmp);
 
-#if OPENSSL_VERSION_NUMBER>=0x0090800fL
     session_id=SSL_SESSION_get_id(sess, &session_id_length);
-#else
-    session_id=(const unsigned char *)sess->session_id;
-    session_id_length=sess->session_id_length;
-#endif
     cache_transfer(SSL_get_SSL_CTX(ssl), CACHE_CMD_NEW,
         SSL_SESSION_get_timeout(sess),
         session_id, session_id_length, val, (size_t)val_len, NULL, NULL);
@@ -972,9 +1099,9 @@ NOEXPORT SSL_SESSION *cache_get(SSL *ssl,
         return NULL;
     val_tmp=val;
     sess=d2i_SSL_SESSION(NULL,
-#if OPENSSL_VERSION_NUMBER>=0x0090800fL
+#if OPENSSL_VERSION_NUMBER>=0x0090707fL
         (const unsigned char **)
-#endif /* OpenSSL version >= 0.8.0 */
+#endif /* OpenSSL version >= 0.9.7g */
         &val_tmp, (long)val_len);
     str_free(val);
     return sess;
@@ -984,12 +1111,7 @@ NOEXPORT void cache_remove(SSL_CTX *ctx, SSL_SESSION *sess) {
     const unsigned char *session_id;
     unsigned int session_id_length;
 
-#if OPENSSL_VERSION_NUMBER>=0x0090800fL
     session_id=SSL_SESSION_get_id(sess, &session_id_length);
-#else
-    session_id=(const unsigned char *)sess->session_id;
-    session_id_length=sess->session_id_length;
-#endif
     cache_transfer(ctx, CACHE_CMD_REMOVE, 0,
         session_id, session_id_length, NULL, 0, NULL, NULL);
 }
@@ -1008,9 +1130,7 @@ NOEXPORT void cache_transfer(SSL_CTX *ctx, const u_char type,
         const u_char *val, const size_t val_len,
         unsigned char **ret, size_t *ret_len) {
     char session_id_txt[2*SSL_MAX_SSL_SESSION_ID_LENGTH+1];
-    const char hex[16]="0123456789ABCDEF";
     const char *type_description[]={"new", "get", "remove"};
-    unsigned i;
     SOCKET s;
     ssize_t len;
     struct timeval t;
@@ -1021,11 +1141,7 @@ NOEXPORT void cache_transfer(SSL_CTX *ctx, const u_char type,
         *ret=NULL;
 
     /* log the request information */
-    for(i=0; i<key_len && i<SSL_MAX_SSL_SESSION_ID_LENGTH; ++i) {
-        session_id_txt[2*i]=hex[key[i]>>4];
-        session_id_txt[2*i+1]=hex[key[i]&0x0f];
-    }
-    session_id_txt[2*i]='\0';
+    bin2hexstring(key, key_len, session_id_txt, sizeof session_id_txt);
     s_log(LOG_INFO,
         "cache_transfer: request=%s, timeout=%ld, id=%s, length=%lu",
         type_description[type], timeout, session_id_txt, (long unsigned)val_len);
@@ -1131,7 +1247,11 @@ NOEXPORT void info_callback(const SSL *ssl, int where, int ret) {
 
     c=SSL_get_ex_data((SSL *)ssl, index_ssl_cli);
     if(c) {
+#if OPENSSL_VERSION_NUMBER>=0x10100000L
+        OSSL_HANDSHAKE_STATE state=SSL_get_state(ssl);
+#else
         int state=SSL_get_state((SSL *)ssl);
+#endif
 
 #if 0
         s_log(LOG_DEBUG, "state = %x", state);
@@ -1143,13 +1263,13 @@ NOEXPORT void info_callback(const SSL *ssl, int where, int ret) {
 #else
         if(state==SSL3_ST_CR_CERT_REQ_A)
 #endif
-            print_client_CA_list(SSL_get_client_CA_list(ssl));
+            print_client_CA_list(SSL_get_client_CA_list((SSL *)ssl));
 #ifndef SSL3_ST_CR_SRVR_DONE_A
         if(state==TLS_ST_CR_SRVR_DONE)
 #else
         if(state==SSL3_ST_CR_SRVR_DONE_A)
 #endif
-            if(!SSL_get_client_CA_list(ssl))
+            if(!SSL_get_client_CA_list((SSL *)ssl))
                 s_log(LOG_INFO, "Client certificate not requested");
 
         /* prevent renegotiation DoS attack */
