@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2017 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -106,6 +106,11 @@ void resolver_init() {
 }
 
 #if defined(USE_WIN32) && !defined(_WIN32_WCE)
+#ifdef __GNUC__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wpragmas"
+#pragma GCC diagnostic ignored "-Wcast-function-type"
+#endif /* __GNUC__ */
 NOEXPORT int get_ipv6(LPTSTR file) {
     HINSTANCE handle;
 
@@ -124,6 +129,9 @@ NOEXPORT int get_ipv6(LPTSTR file) {
     }
     return 1; /* IPv6 detected -> OK */
 }
+#ifdef __GNUC__
+#pragma GCC diagnostic pop
+#endif /* __GNUC__ */
 #endif
 
 /**************************************** stunnel resolver API */
@@ -138,8 +146,6 @@ unsigned name2addr(SOCKADDR_UNION *addr, char *name, int passive) {
     if(retval)
         addrlist2addr(addr, addr_list);
     str_free(addr_list->addr);
-    str_free(addr_list->fd);
-    str_free(addr_list->session);
     str_free(addr_list);
     return retval;
 }
@@ -155,8 +161,6 @@ unsigned hostport2addr(SOCKADDR_UNION *addr,
     if(num)
         addrlist2addr(addr, addr_list);
     str_free(addr_list->addr);
-    str_free(addr_list->fd);
-    str_free(addr_list->session);
     str_free(addr_list);
     return num;
 }
@@ -194,16 +198,10 @@ unsigned name2addrlist(SOCKADDR_LIST *addr_list, char *name) {
             s_log(LOG_ERR, "Unix socket path is too long");
             return 0; /* no results */
         }
-        addr_list->addr=str_realloc(addr_list->addr,
+        addr_list->addr=str_realloc_detached(addr_list->addr,
             (addr_list->num+1)*sizeof(SOCKADDR_UNION));
         addr_list->addr[addr_list->num].un.sun_family=AF_UNIX;
         strcpy(addr_list->addr[addr_list->num].un.sun_path, name);
-        addr_list->fd=str_realloc(addr_list->fd,
-            (addr_list->num+1)*sizeof(SOCKET));
-        addr_list->fd[addr_list->num]=INVALID_SOCKET;
-        addr_list->session=str_realloc(addr_list->session,
-            (addr_list->num+1)*sizeof(SSL_SESSION *));
-        addr_list->session[addr_list->num]=NULL;
         ++(addr_list->num);
         return 1; /* ok - return the number of new addresses */
     }
@@ -228,9 +226,9 @@ unsigned name2addrlist(SOCKADDR_LIST *addr_list, char *name) {
 
 unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
         char *host_name, char *port_name) {
-    struct addrinfo hints, *res=NULL, *cur;
+    struct addrinfo hints, *res, *cur;
     int err, retry=0;
-    unsigned num=0;
+    unsigned num;
 
     memset(&hints, 0, sizeof hints);
 #if defined(USE_IPv6) || defined(USE_WIN32)
@@ -241,19 +239,20 @@ unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
     hints.ai_socktype=SOCK_STREAM;
     hints.ai_protocol=IPPROTO_TCP;
     hints.ai_flags=0;
-    if(addr_list->passive) {
-        hints.ai_family=AF_INET; /* first try IPv4 for passive requests */
+    if(addr_list->passive)
         hints.ai_flags|=AI_PASSIVE;
-    }
 #ifdef AI_ADDRCONFIG
     hints.ai_flags|=AI_ADDRCONFIG;
 #endif
     for(;;) {
+        res=NULL;
         err=getaddrinfo(host_name, port_name, &hints, &res);
-        if(!err)
+        if(!err) /* success */
             break;
-        if(res)
-            freeaddrinfo(res);
+        if(err==EAI_SERVICE) {
+            s_log(LOG_ERR, "Unknown TCP service \"%s\"", port_name);
+            return 0; /* error */
+        }
         if(err==EAI_AGAIN && ++retry<=3) {
             s_log(LOG_DEBUG, "getaddrinfo: EAI_AGAIN received: retrying");
             sleep(1);
@@ -265,19 +264,6 @@ unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
             continue; /* retry for unconfigured network interfaces */
         }
 #endif
-#if defined(USE_IPv6) || defined(USE_WIN32)
-        if(hints.ai_family==AF_INET) {
-            hints.ai_family=AF_UNSPEC;
-            continue; /* retry for non-IPv4 addresses */
-        }
-#endif
-        break;
-    }
-    if(err==EAI_SERVICE) {
-        s_log(LOG_ERR, "Unknown TCP service \"%s\"", port_name);
-        return 0; /* error */
-    }
-    if(err) {
         s_log(LOG_ERR, "Error resolving \"%s\": %s",
             host_name ? host_name :
                 (addr_list->passive ? DEFAULT_ANY : DEFAULT_LOOPBACK),
@@ -285,26 +271,24 @@ unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
         return 0; /* error */
     }
 
-    /* copy the list of addresses */
+    /* find the number of newly resolved addresses */
+    num=0;
     for(cur=res; cur; cur=cur->ai_next) {
         if(cur->ai_addrlen>(int)sizeof(SOCKADDR_UNION)) {
             s_log(LOG_ERR, "INTERNAL ERROR: ai_addrlen value too big");
             freeaddrinfo(res);
             return 0; /* no results */
         }
-        addr_list->addr=str_realloc(addr_list->addr,
-            (addr_list->num+1)*sizeof(SOCKADDR_UNION));
-        memcpy(&addr_list->addr[addr_list->num], cur->ai_addr,
-            (size_t)cur->ai_addrlen);
-        addr_list->fd=str_realloc(addr_list->fd,
-            (addr_list->num+1)*sizeof(SOCKET));
-        addr_list->fd[addr_list->num]=INVALID_SOCKET;
-        addr_list->session=str_realloc(addr_list->session,
-            (addr_list->num+1)*sizeof(SSL_SESSION *));
-        addr_list->session[addr_list->num]=NULL;
-        ++(addr_list->num);
         ++num;
     }
+
+    /* append the newly resolved addresses to addr_list->addr */
+    addr_list->addr=str_realloc_detached(addr_list->addr,
+        (addr_list->num+num)*sizeof(SOCKADDR_UNION));
+    for(cur=res; cur; cur=cur->ai_next)
+        memcpy(&addr_list->addr[(addr_list->num)++], cur->ai_addr,
+            (size_t)cur->ai_addrlen);
+
     freeaddrinfo(res);
     return num; /* ok - return the number of new addresses */
 }
@@ -320,8 +304,6 @@ void addrlist_clear(SOCKADDR_LIST *addr_list, int passive) {
 NOEXPORT void addrlist_reset(SOCKADDR_LIST *addr_list) {
     addr_list->num=0;
     addr_list->addr=NULL;
-    addr_list->fd=NULL;
-    addr_list->session=NULL;
     addr_list->start=0;
     addr_list->parent=addr_list; /* allow a copy to locate its parent */
 }
@@ -329,13 +311,11 @@ NOEXPORT void addrlist_reset(SOCKADDR_LIST *addr_list) {
 unsigned addrlist_dup(SOCKADDR_LIST *dst, const SOCKADDR_LIST *src) {
     memcpy(dst, src, sizeof(SOCKADDR_LIST));
     if(src->num) { /* already resolved */
-        dst->addr=str_alloc(src->num*sizeof(SOCKADDR_UNION));
+        dst->addr=str_alloc_detached(src->num*sizeof(SOCKADDR_UNION));
         memcpy(dst->addr, src->addr, src->num*sizeof(SOCKADDR_UNION));
     } else { /* delayed resolver */
         addrlist_resolve(dst);
     }
-    /* a client does not have its own local copy of
-       src->session and src->fd */
     return dst->num;
 }
 
@@ -459,7 +439,7 @@ NOEXPORT int getaddrinfo(const char *node, const char *service,
     /* not numerical: need to call resolver library */
     *res=NULL;
     ai=NULL;
-    stunnel_write_lock(&stunnel_locks[LOCK_INET]);
+    CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_INET]);
 #ifdef HAVE_GETHOSTBYNAME2
     h=gethostbyname2(node, AF_INET6);
     if(h) /* some IPv6 addresses found */
@@ -477,7 +457,7 @@ NOEXPORT int getaddrinfo(const char *node, const char *service,
 #ifdef HAVE_ENDHOSTENT
     endhostent();
 #endif
-    stunnel_write_unlock(&stunnel_locks[LOCK_INET]);
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_INET]);
     if(retval) { /* error: free allocated memory */
         freeaddrinfo(*res);
         *res=NULL;
@@ -612,10 +592,11 @@ int getnameinfo(const struct sockaddr *sa, socklen_t salen,
                 (void *)&((struct sockaddr_in *)sa)->sin_addr,
             host, hostlen);
 #else /* USE_IPv6 */
-        stunnel_write_lock(&stunnel_locks[LOCK_INET]); /* inet_ntoa is not mt-safe */
+        /* inet_ntoa is not mt-safe */
+        CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_INET]);
         strncpy(host, inet_ntoa(((struct sockaddr_in *)sa)->sin_addr),
             hostlen);
-        stunnel_write_unlock(&stunnel_locks[LOCK_INET]);
+        CRYPTO_THREAD_unlock(stunnel_locks[LOCK_INET]);
         host[hostlen-1]='\0';
 #endif /* USE_IPv6 */
     }
